#include "header/client.h"
#include "header/vr_ovr.h"

#define CVAR_CLIENT CVAR_ARCHIVE
#define SCR_CenterAlert SCR_CenterPrint
extern float    ClampCvar( float min, float max, float value );

void Cvar_SetInteger( const char * var_name, int32_t integer )
{
	char val[32];
	Com_sprintf( val, sizeof( val ), "%i", integer );
	Cvar_Set( var_name, val );
}

// vr_rift.c

void IN_DownDown( void );
void IN_DownUp( void );

#ifdef FIXED_GAMEPADS
static struct
{
	int32_t repeatCount;
	int32_t lastSendTime;
} gamepad_repeatstatus[K_GAMEPAD_RIGHT - K_GAMEPAD_LSTICK_UP + 1];
#endif

static const int32_t InitialRepeatDelay = 220;
static const int32_t RepeatDelay        = 160;

void VR_Rift_GetHMDResolution( int32_t * width, int32_t * height );
void VR_Rift_FrameStart();
void VR_Rift_FrameEnd();

int32_t VR_Rift_Enable();
void    VR_Rift_Disable();
int32_t VR_Rift_Init();
ovrBool VR_Rift_InitSensor();
void    VR_Rift_Shutdown();
int32_t VR_Rift_getOrientation( float euler[3] );
void    VR_Rift_ResetHMDOrientation();
int32_t VR_Rift_SetPredictionTime( float time );
int32_t VR_Rift_getPosition( float pos[3] );
int32_t VR_Rift_GetControllerOrientation( float euler[3] );
void    VR_Rift_GetInputCommands();
void    VR_Rift_GetControllerMoveInput( usercmd_t * cmd );

cvar_t *        vr_rift_debug;
cvar_t *        vr_rift_maxfov;
cvar_t *        vr_rift_enable;
cvar_t *        vr_rift_autoprediction;
cvar_t *        vr_rift_dk2_color_hack;
cvar_t *        vr_rift_trackingloss;
static cvar_t * gamepad_trigger_threshold;
static cvar_t * gamepad_stick_mode;
static cvar_t * gamepad_stick_toggle;

ovrSession       riftHmd;
ovrHmdDesc       hmdDesc;
ovrGraphicsLuid  graphicsLuid;
ovrEyeRenderDesc eyeDesc[2];

ovrTrackingState trackingState;
ovrInputState    lastInputState;
static qboolean  gamepad_sticktoggle[2];
double           frameTime;
long long        riftFrameIndex;
static ovrBool   sensorEnabled         = 0;
static float     ovrPlayerHeightMeters = 0;
static qboolean  isAutoCrouching = false, wasAutoCrouching = false;

static ovrBool libovrInitialized = 0;

static double prediction_time;

rift_render_export_t renderExport;

hmd_interface_t hmd_rift = {HMD_RIFT,
                            VR_Rift_Init,
                            VR_Rift_Shutdown,
                            VR_Rift_Enable,
                            VR_Rift_Disable,
                            VR_Rift_FrameStart,
                            VR_Rift_FrameEnd,
                            VR_Rift_ResetHMDOrientation,
                            VR_Rift_getOrientation,
                            VR_Rift_getPosition,
                            VR_Rift_SetPredictionTime,
                            NULL,
                            VR_Rift_GetHMDResolution,
                            VR_Rift_GetControllerOrientation,
                            VR_Rift_GetInputCommands,
                            VR_Rift_GetControllerMoveInput};

static riftname_t hmdnames[255];

void VR_Rift_GetFOV( float * fovx, float * fovy )
{
	*fovx = hmdDesc.DefaultEyeFov[ovrEye_Left].LeftTan + hmdDesc.DefaultEyeFov[ovrEye_Left].RightTan;
	*fovy = hmdDesc.DefaultEyeFov[ovrEye_Left].UpTan + hmdDesc.DefaultEyeFov[ovrEye_Left].DownTan;
}

void VR_Rift_GetHMDResolution( int32_t * width, int32_t * height )
{
	if( riftHmd )
	{
		*width  = hmdDesc.Resolution.w;
		*height = hmdDesc.Resolution.h;
	}
}

void VR_Rift_QuatToEuler( ovrQuatf q, vec3_t e )
{
	vec_t w    = q.w;
	vec_t Q[3] = {q.x, q.y, q.z};

	vec_t ww  = w * w;
	vec_t Q11 = Q[1] * Q[1];
	vec_t Q22 = Q[0] * Q[0];
	vec_t Q33 = Q[2] * Q[2];

	vec_t s2 = -2.0f * ( -w * Q[0] + Q[1] * Q[2] );

	if( s2 < -1.0f + 0.0000001f )
	{ // South pole singularity
		e[YAW]   = 0.0f;
		e[PITCH] = -M_PI / 2.0f;
		e[ROLL]  = atan2( 2.0f * ( -Q[1] * Q[0] + w * Q[2] ), ww + Q22 - Q11 - Q33 );
	}
	else if( s2 > 1.0f - 0.0000001f )
	{ // North pole singularity
		e[YAW]   = 0.0f;
		e[PITCH] = M_PI / 2.0f;
		e[ROLL]  = atan2( 2.0f * ( -Q[1] * Q[0] + w * Q[2] ), ww + Q22 - Q11 - Q33 );
	}
	else
	{
		e[YAW]   = -atan2( -2.0f * ( w * Q[1] - -Q[0] * Q[2] ), ww + Q33 - Q11 - Q22 );
		e[PITCH] = asin( s2 );
		e[ROLL]  = atan2( 2.0f * ( w * Q[2] - -Q[1] * Q[0] ), ww + Q11 - Q22 - Q33 );
	}

	e[PITCH] = -RAD2DEG( e[PITCH] );
	e[YAW]   = RAD2DEG( e[YAW] );
	e[ROLL]  = -RAD2DEG( e[ROLL] );
}

int32_t VR_Rift_getOrientation( float euler[3] )
{
	double           time = 0.0;
	ovrSessionStatus sessionStatus;

	if( !riftHmd ) return 0;

	ovr_GetSessionStatus( riftHmd, &sessionStatus );

	if( sessionStatus.ShouldRecenter ) { ovr_RecenterTrackingOrigin( riftHmd ); }

	if( vr_rift_autoprediction->value > 0 )
		time = frameTime;
	else
		time = ovr_GetTimeInSeconds() + prediction_time;
	trackingState = ovr_GetTrackingState( riftHmd, time, true );
	if( trackingState.StatusFlags & ovrStatus_OrientationTracked )
	{
		// vec3_t temp;
		VR_Rift_QuatToEuler( trackingState.HeadPose.ThePose.Orientation, euler );
		/*if (trackingState.StatusFlags & ovrStatus_PositionTracked) {
		    VR_Rift_QuatToEuler(trackingState.LeveledCameraPose.Orientation, temp);
		    renderExport.cameraYaw = euler[YAW] - temp[YAW];
		    AngleClamp(&renderExport.cameraYaw);
		}
		else*/
		{
			renderExport.cameraYaw = 0.0;
		}
		return 1;
	}
	return 0;
}

int32_t VR_Rift_GetControllerOrientation( float euler[3] )
{
	double  time   = 0.0;
	int32_t vrHand = 0;

	if( !riftHmd ) return 0;

	if( vr_controllermode->value ) { vrHand = (int32_t)vr_controllermode->value; }
	if( vrHand == 0 ) { return 0; }
	vrHand = vrHand == 2 ? ovrHand_Left : ovrHand_Right;

	if( trackingState.HandStatusFlags[vrHand] & ovrStatus_OrientationTracked )
	{
		VR_Rift_QuatToEuler( trackingState.HandPoses[vrHand].ThePose.Orientation, euler );
		euler[ROLL] = -euler[ROLL];
		return 1;
	}

	return 0;
}

void    SCR_CenterAlert( char * str );
int32_t VR_Rift_getPosition( float pos[3] )
{
	qboolean tracked = 0;
	if( !riftHmd ) return 0;

	//	if (!sensorEnabled)
	//		VR_Rift_InitSensor();

	if( sensorEnabled && renderExport.positionTracked )
	{
		tracked = trackingState.StatusFlags & ovrStatus_PositionTracked ? 1 : 0;
		if( tracked )
		{
			ovrPosef pose = trackingState.HeadPose.ThePose;
			VectorSet( pos, -pose.Position.z, pose.Position.x, pose.Position.y );

			if( vr_autocrouch->value )
			{
				if( pos[2] < ovrPlayerHeightMeters / -5.0 )
				{
					isAutoCrouching = true;
					if( cl.frame.playerstate.pmove.pm_flags & PMF_DUCKED )
					{
						pos[2] += 0.75; // Viewheight for player +22 -> -2 in PM_CheckDuck(). 24*pHeightMeters/pHeightUnits = 0.75m
					}
				}
				else
				{
					isAutoCrouching = false;
				}
			}

			VectorScale( pos, ( PLAYER_HEIGHT_UNITS / PLAYER_HEIGHT_M ), pos );
		}

		if( vr_rift_trackingloss->value > 0 && trackingState.StatusFlags & ovrTracker_Connected )
		{
			if( tracked && !renderExport.hasPositionLock && vr_rift_debug->value )
				SCR_CenterAlert( "Position tracking enabled" );
			else if( !tracked && renderExport.hasPositionLock && vr_rift_debug->value )
				SCR_CenterAlert( "Position tracking interrupted" );
		}

		renderExport.hasPositionLock = tracked;
	}
	return tracked;
}

void VR_Rift_ResetHMDOrientation()
{
	if( riftHmd )
	{
		ovrTrackingState state;

		ovr_RecenterTrackingOrigin( riftHmd );

		ovr_SetTrackingOriginType( riftHmd, ovrTrackingOrigin_FloorLevel );
		state = ovr_GetTrackingState( riftHmd, 0, false );
		ovr_SetTrackingOriginType( riftHmd, ovrTrackingOrigin_EyeLevel );
		ovrPlayerHeightMeters = state.HeadPose.ThePose.Position.y;
	}
}

void Rift_ParseThumbstick( ovrVector2f thumbstick, vec3_t out )
{
	float LX = thumbstick.x, LY = thumbstick.y;
	// determine how far the controller is pushed
	float mag = sqrt( LX * LX + LY * LY );

	// determine the direction the controller is pushed
	if( mag > 0 )
	{
		out[0] = LX / mag;
		out[1] = LY / mag;
	}
	else
	{
		out[0] = 0;
		out[1] = 0;
	}
	out[2] = mag;
}

void VR_Rift_GetControllerMoveInput( usercmd_t * cmd )
{
	vec3_t leftPos, rightPos;

	Rift_ParseThumbstick( lastInputState.Thumbstick[ovrHand_Left], leftPos );
	Rift_ParseThumbstick( lastInputState.Thumbstick[ovrHand_Right], rightPos );

//#ifdef FIXED_GAMEPADS
	ThumbstickMove( cmd, leftPos, rightPos, true );
//#endif
}

#ifdef FIXED_GAMEPADS
static void RiftTouchInput_HandleRepeat( keynum_t key, int32_t delayMultiplier )
{
	int32_t  index = key - K_GAMEPAD_LSTICK_UP;
	int32_t  delay = 0;
	qboolean send  = false;

	if( index < 0 || index > ( K_GAMEPAD_RIGHT - K_GAMEPAD_LSTICK_UP ) ) return;

	if( gamepad_repeatstatus[index].repeatCount == 0 )
		delay = InitialRepeatDelay * delayMultiplier;
	else
		delay = RepeatDelay * delayMultiplier;

	if( cl.time > gamepad_repeatstatus[index].lastSendTime + delay || cl.time < gamepad_repeatstatus[index].lastSendTime ) send = true;

	if( send )
	{
		gamepad_repeatstatus[index].lastSendTime = cl.time;
		gamepad_repeatstatus[index].repeatCount++;
		Key_Event( key, true, 0 );
	}
}
#endif

void VR_Rift_GetInputCommands()
{
	// NOTE: Gamepad / XBox controller isn't handled here because Q2 already supports & handles XInput devices separately
	int32_t i;
	int32_t handleThumbStickButtonsNormally = !gamepad_stick_toggle->value || cls.key_dest == key_menu;

	if( riftHmd && vr_autocrouch->value )
	{
		if( !wasAutoCrouching && isAutoCrouching ) { Cmd_ExecuteString( "+movedown" ); }
		if( wasAutoCrouching && !isAutoCrouching ) { Cmd_ExecuteString( "-movedown" ); }
		wasAutoCrouching = isAutoCrouching;
	}

	if( riftHmd && ( ovr_GetConnectedControllerTypes( riftHmd ) & ovrControllerType_Touch ) )
	{
		ovrInputState inputState;
		const float   sq1over2 = sqrt( 0.5 );
		if( OVR_SUCCESS( ovr_GetInputState( riftHmd, ovrControllerType_Touch, &inputState ) ) )
		{
			{
				qboolean leftTriggerOld, leftTrigger, rightTriggerOld, rightTrigger, leftGripOld, leftGrip, rightGripOld, rightGrip;
				float    triggerThreshold = ClampCvar( 0.04, 0.96, gamepad_trigger_threshold->value );

				leftTriggerOld = lastInputState.IndexTrigger[ovrHand_Left] > triggerThreshold;
				leftTrigger    = inputState.IndexTrigger[ovrHand_Left] > triggerThreshold;
#ifdef FIXED_GAMEPADS
				if( !leftTriggerOld && leftTrigger ) Key_Event( K_GAMEPAD_LT, true, 0 );
				if( leftTriggerOld && !leftTrigger ) Key_Event( K_GAMEPAD_LT, false, 0 );
#endif

				rightTriggerOld = lastInputState.IndexTrigger[ovrHand_Right] > triggerThreshold;
				rightTrigger    = inputState.IndexTrigger[ovrHand_Right] > triggerThreshold;
#ifdef FIXED_GAMEPADS
				if( !rightTriggerOld && rightTrigger ) Key_Event( K_GAMEPAD_RT, true, 0 );
				if( rightTriggerOld && !rightTrigger ) Key_Event( K_GAMEPAD_RT, false, 0 );
#endif

				leftGripOld = lastInputState.HandTrigger[ovrHand_Left] > triggerThreshold;
				leftGrip    = inputState.HandTrigger[ovrHand_Left] > triggerThreshold;
#ifdef FIXED_GAMEPADS
				if( !leftGripOld && leftGrip ) Key_Event( K_GAMEPAD_LS, true, 0 );
				if( leftGripOld && !leftGrip ) Key_Event( K_GAMEPAD_LS, false, 0 );
#endif

				rightGripOld = lastInputState.HandTrigger[ovrHand_Right] > triggerThreshold;
				rightGrip    = inputState.HandTrigger[ovrHand_Right] > triggerThreshold;
#ifdef FIXED_GAMEPADS
				if( !rightGripOld && rightGrip ) Key_Event( K_GAMEPAD_RS, true, 0 );
				if( rightGripOld && !rightGrip ) Key_Event( K_GAMEPAD_RS, false, 0 );
#endif
			}

			{
				int32_t aimStickHand = (int32_t)gamepad_stick_mode->value == 1 ? ovrHand_Left : ovrHand_Right;
				if( cls.key_dest != key_menu && vr_controllermode->value && ( trackingState.HandStatusFlags[aimStickHand] & ovrStatus_OrientationTracked ) )
				{
					// For Oculus Touch, interpret aim/look thumbstick up/ down as gamepad up/down when VR controller tracking is on.
					// This is because pitch (aim up/ down) is tied to the controller, so thumbstick up/down would otherwise be unused.
					// And Oculus Touch has no DPad, so the user has fewer inputs to work with. This way, with the default bindings,
					// the user can access Next Item & Use Item when using VR controller input.
					vec3_t  oldStick, newStick;
					int32_t oldDir = 0, newDir = 0;
					Rift_ParseThumbstick( lastInputState.Thumbstick[aimStickHand], oldStick );
					Rift_ParseThumbstick( inputState.Thumbstick[aimStickHand], newStick );
#ifdef FIXED_GAMEPADS
					if( oldStick[1] * oldStick[2] <= -sq1over2 ) oldDir = K_GAMEPAD_DOWN;
					if( oldStick[1] * oldStick[2] >= sq1over2 ) oldDir = K_GAMEPAD_UP;
					if( newStick[1] * newStick[2] <= -sq1over2 ) newDir = K_GAMEPAD_DOWN;
					if( newStick[1] * newStick[2] >= sq1over2 ) newDir = K_GAMEPAD_UP;
#endif
					if( newDir != oldDir )
					{
						if( oldDir != 0 ) { Key_Event( oldDir, false, 0 ); }
#ifdef FIXED_GAMEPADS
						if( newDir != 0 ) { RiftTouchInput_HandleRepeat( newDir, 3 ); }
#endif
					}
				}
			}

			if( cls.key_dest == key_menu )
			{
				for( i = 0; i < 2; i++ )
				{
					vec3_t  oldStick, newStick;
					int32_t oldDir = 0, newDir = 0;
#ifdef FIXED_GAMEPADS
					int32_t upKey = i == 0 ? K_GAMEPAD_LSTICK_UP : K_GAMEPAD_RSTICK_UP;
#endif
					Rift_ParseThumbstick( lastInputState.Thumbstick[i], oldStick );
					Rift_ParseThumbstick( inputState.Thumbstick[i], newStick );
#ifdef FIXED_GAMEPADS
					if( oldStick[1] * oldStick[2] <= -sq1over2 ) oldDir = upKey + 1; // Down
					if( oldStick[1] * oldStick[2] >= sq1over2 ) oldDir = upKey;      // Up
					if( oldStick[0] * oldStick[2] <= -sq1over2 ) oldDir = upKey + 2; // Left
					if( oldStick[0] * oldStick[2] >= sq1over2 ) oldDir = upKey + 3;  // Right
					if( newStick[1] * newStick[2] <= -sq1over2 ) newDir = upKey + 1;
					if( newStick[1] * newStick[2] >= sq1over2 ) newDir = upKey;
					if( newStick[0] * newStick[2] <= -sq1over2 ) newDir = upKey + 2;
					if( newStick[0] * newStick[2] >= sq1over2 ) newDir = upKey + 3;
#endif
					if( newDir != oldDir )
					{
						if( oldDir != 0 ) { Key_Event( oldDir, false, 0 ); }
#ifdef FIXED_GAMEPADS
						if( newDir != 0 ) { RiftTouchInput_HandleRepeat( newDir, 1 ); }
#endif
					}
				}
			}

			{
				int32_t   TouchButtonCount = handleThumbStickButtonsNormally ? 7 : 5;
				ovrButton ovrButtons[]     = {
                      ovrButton_A, ovrButton_B, ovrButton_X, ovrButton_Y, ovrButton_Enter, ovrButton_LThumb, ovrButton_RThumb,
                };
#ifdef FIXED_GAMEPADS
				int32_t gamepadButtons[] = {K_GAMEPAD_A, K_GAMEPAD_B, K_GAMEPAD_X, K_GAMEPAD_Y, K_GAMEPAD_START, K_GAMEPAD_LEFT_STICK, K_GAMEPAD_RIGHT_STICK};
#endif
				for( i = 0; i < TouchButtonCount; i++ )
				{
					qboolean oldState, newState;
					oldState = lastInputState.Buttons & ovrButtons[i];
					newState = inputState.Buttons & ovrButtons[i];
#ifdef FIXED_GAMEPADS
					if( !oldState && newState ) Key_Event( gamepadButtons[i], true, 0 );
					if( oldState && !newState ) Key_Event( gamepadButtons[i], false, 0 );
#endif
				}
			}

			if( handleThumbStickButtonsNormally ) { memset( gamepad_sticktoggle, 0, sizeof( gamepad_sticktoggle ) ); }
			else
			{
				if( ( inputState.Buttons & ovrButton_LThumb ) && !( lastInputState.Buttons & ovrButton_LThumb ) )
				{
					gamepad_sticktoggle[0] = !gamepad_sticktoggle[0];
#ifdef FIXED_GAMEPADS
					Key_Event( K_GAMEPAD_LEFT_STICK, gamepad_sticktoggle[0], 0 );
#endif
				}
				if( ( inputState.Buttons & ovrButton_RThumb ) && !( lastInputState.Buttons & ovrButton_RThumb ) )
				{
					gamepad_sticktoggle[1] = !gamepad_sticktoggle[1];
#ifdef FIXED_GAMEPADS
					Key_Event( K_GAMEPAD_RIGHT_STICK, gamepad_sticktoggle[1], 0 );
#endif
				}
			}

			memcpy( &lastInputState, &inputState, sizeof( ovrInputState ) );
			return;
		}
	}
	memset( &lastInputState, 0, sizeof( ovrInputState ) );
}

ovrBool VR_Rift_InitSensor()
{
	sensorEnabled = hmdDesc.AvailableTrackingCaps & ovrTrackingCap_Position;
	return hmdDesc.AvailableTrackingCaps & ovrTrackingCap_Orientation;
}

int32_t VR_Rift_SetPredictionTime( float time )
{
	prediction_time = time / 1000.0;
	if( vr_rift_debug->value ) Com_Printf( "VR_Rift: Set HMD Prediction time to %.1fms\n", time );
	return 1;
}

void VR_Rift_FrameStart()
{
	if( !renderExport.withinFrame ) { frameTime = ovr_GetPredictedDisplayTime( riftHmd, riftFrameIndex ); }
	renderExport.withinFrame = true;
}

void VR_Rift_FrameEnd() { renderExport.withinFrame = false; }

float VR_Rift_GetGammaMin()
{
	float result = 0.0;
	if( riftHmd && hmdDesc.Type >= ovrHmd_DK2 ) result = 1.0 / 255.0;
	return result;
}

riftname_t * VR_Rift_GetNameList() { return hmdnames; }

int32_t VR_Rift_Enable()
{
	qboolean  isDebug = false;
	uint32_t  device  = 0;
	ovrResult result  = 0;

	if( !vr_rift_enable->value ) return 0;

	if( !libovrInitialized )
	{
		result = ovr_Initialize( 0 );

		if( OVR_FAILURE( result ) )
		{
			Com_Printf( "VR_Rift: Fatal error: could not initialize LibOVR!\n" );
			libovrInitialized = 0;
			return 0;
		}
		else
		{
			Com_Printf( "VR_Rift: %s initialized...\n", ovr_GetVersionString() );
			libovrInitialized = 1;
		}
	}
	{
		qboolean found = false;

		Com_Printf( "VR_Rift: Enumerating devices...\n" );

		Com_Printf( "VR_Rift: Initializing HMD: " );

		renderExport.withinFrame = false;
		result                   = ovr_Create( &riftHmd, &graphicsLuid );

		if( OVR_FAILURE( result ) ) { Com_Printf( "no HMD detected!\n" ); }
		else
		{
			Com_Printf( "ok!\n" );
		}

		hmdDesc = ovr_GetHmdDesc( riftHmd );

		if( hmdDesc.AvailableTrackingCaps & ovrTrackingCap_Position )
		{
			renderExport.positionTracked = ( qboolean ) true;
			Com_Printf( "...supports position tracking\n" );
		}
		else
		{
			renderExport.positionTracked = ( qboolean ) false;
		}
		Com_Printf( "...has type %s\n", hmdDesc.ProductName );
		Com_Printf( "...has %ux%u native resolution\n", hmdDesc.Resolution.w, hmdDesc.Resolution.h );

		if( !VR_Rift_InitSensor() ) { Com_Printf( "VR_Rift: Sensor initialization failed!\n" ); }
		return 1;
	}
}

void VR_Rift_Disable()
{
	renderExport.withinFrame = false;
	if( riftHmd ) { ovr_Destroy( riftHmd ); }
	riftHmd = NULL;
}

int32_t VR_Rift_Init()
{
	/*
	if (!libovrInitialized)
	{
	libovrInitialized = ovr_Initialize();
	if (!libovrInitialized)
	{
	Com_Printf("VR_Rift: Fatal error: could not initialize LibOVR!\n");
	return 0;
	} else {
	Com_Printf("VR_Rift: %s initialized...\n",ovr_GetVersionString());
	}
	}
	*/

	vr_rift_trackingloss      = Cvar_Get( "vr_rift_trackingloss", "0", CVAR_CLIENT );
	vr_rift_maxfov            = Cvar_Get( "vr_rift_maxfov", "0", CVAR_CLIENT );
	vr_rift_enable            = Cvar_Get( "vr_rift_enable", "1", CVAR_CLIENT );
	vr_rift_dk2_color_hack    = Cvar_Get( "vr_rift_dk2_color_hack", "0", CVAR_CLIENT );
	vr_rift_debug             = Cvar_Get( "vr_rift_debug", "0", CVAR_CLIENT );
	vr_rift_autoprediction    = Cvar_Get( "vr_rift_autoprediction", "1", CVAR_CLIENT );
	gamepad_trigger_threshold = Cvar_Get( "gamepad_trigger_threshold", "0.12", CVAR_ARCHIVE );
	gamepad_stick_toggle      = Cvar_Get( "gamepad_stick_toggle", "0", CVAR_ARCHIVE );
	gamepad_stick_mode        = Cvar_Get( "gamepad_stick_mode", "0", CVAR_ARCHIVE );

	memset( &lastInputState, 0, sizeof( ovrInputState ) );
	riftFrameIndex        = 0;
	ovrPlayerHeightMeters = 1.75;
	isAutoCrouching       = false;
	wasAutoCrouching      = false;

	return 1;
}

void VR_Rift_Shutdown()
{
	if( libovrInitialized ) ovr_Shutdown();
	libovrInitialized = 0;
}

// vr_main.c

cvar_t * vr_enabled;
cvar_t * vr_autoenable;
cvar_t * vr_autoipd;
cvar_t * vr_autofov;
cvar_t * vr_autofov_scale;
cvar_t * vr_ipd;
cvar_t * vr_hud_fov;
cvar_t * vr_hud_depth;
cvar_t * vr_hud_transparency;
cvar_t * vr_hud_segments;
cvar_t * vr_hud_width;
cvar_t * vr_hud_height;
cvar_t * vr_chromatic;
cvar_t * vr_aimlaser;
cvar_t * vr_aimmode;
cvar_t * vr_aimmode_deadzone_yaw;
cvar_t * vr_aimmode_deadzone_pitch;
cvar_t * vr_controllermode;
cvar_t * vr_supersampling;
cvar_t * vr_comfortturn;
cvar_t * vr_viewmove;
cvar_t * vr_walkspeed;
cvar_t * vr_hud_bounce;
cvar_t * vr_hud_bounce_falloff;
cvar_t * vr_hud_deadzone_yaw;
cvar_t * vr_hud_fixed;
cvar_t * vr_nosleep;
cvar_t * vr_neckmodel;
cvar_t * vr_neckmodel_up;
cvar_t * vr_neckmodel_forward;
cvar_t * vr_heightoffset;
cvar_t * vr_autocrouch;
cvar_t * vr_hmdtype;
cvar_t * vr_prediction;
cvar_t * vr_hmdstring;
cvar_t * vr_positiontracking;
cvar_t * vr_force_fullscreen;
cvar_t * vr_force_resolution;

static vec3_t vr_lastOrientation;
static vec3_t vr_orientation;
static vec4_t vr_smoothSeries;
static vec4_t vr_doubleSmoothSeries;
static vec4_t vr_smoothOrientation;
static vec3_t vr_controllerOrientation, vr_lastControllerOrientation;

static qboolean stale, controllerstale;
float           vr_orientation_deltayaw;
float           vr_comfortturn_delta;

static hmd_interface_t available_hmds[NUM_HMD_TYPES];

static hmd_interface_t * hmd;

static hmd_interface_t hmd_none = {HMD_NONE, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL};

const char * hmd_names[] = {"none", "Oculus Rift",
#ifndef NO_STEAM
                            "SteamVR",
#endif
                            0};

//
// General functions
//

int32_t VR_GetSensorOrientation()
{
	vec3_t euler;
	vec4_t currentPos;

	float emaWeight = 2 / ( vr_hud_bounce_falloff->value + 1 );

	if( !hmd ) return 0;

	if( !hmd->getOrientation || !hmd->getOrientation( euler ) ) VectorSet( euler, 0, 0, 0 );

	EulerToQuat( vr_lastOrientation, currentPos );
	VectorCopy( vr_orientation, vr_lastOrientation );
	VectorCopy( euler, vr_orientation );

	// convert last position to quaternions for SLERP
	//	EulerToQuat(vr_orientation,currentPos);

	// SLERP between previous EMA orientation and new EMA orientation
	//	LerpQuat(vr_smoothSeries, quat, emaWeight, vr_smoothSeries);
	SlerpQuat( vr_smoothSeries, currentPos, emaWeight, vr_smoothSeries );
	SlerpQuat( vr_doubleSmoothSeries, vr_smoothSeries, emaWeight, vr_doubleSmoothSeries );

	if( vr_hud_bounce->value == VR_HUD_BOUNCE_LES )
	{
		vec4_t level, trend, zero = {1, 0, 0, 0};
		vec4_t negDouble;
		// fast conjugate
		QuatCopy( vr_doubleSmoothSeries, negDouble );
		negDouble[0] = -negDouble[0];

		QuatMultiply( vr_smoothSeries, negDouble, trend );
		QuatMultiply( trend, vr_smoothSeries, level );

		SlerpQuat( zero, trend, emaWeight / ( 1 - emaWeight ), trend );
		QuatMultiply( level, trend, vr_smoothOrientation );
	}
	else if( vr_hud_bounce->value <= VR_HUD_BOUNCE_SES )
	{
		QuatCopy( vr_smoothSeries, vr_smoothOrientation );
	}
	return 1;
}

int32_t VR_GetControllerOrientation( vec3_t angle )
{
	vec3_t  euler;
	int32_t result = 0;

	if( !hmd || (int32_t)vr_controllermode->value == 0 ) return 0;

	if( controllerstale )
	{
		if( hmd->getControllerOrientation )
		{
			result = hmd->getControllerOrientation( euler );
			if( result )
			{
				controllerstale = 0;

				VectorCopy( vr_controllerOrientation, vr_lastControllerOrientation );
				VectorCopy( euler, vr_controllerOrientation );
			}
			else
			{
				return 0;
			}
		}
	}

	VectorCopy( vr_controllerOrientation, angle );
	return 1;
}

void VR_GetControllerOrientationDelta( vec3_t angle )
{
	if( !hmd ) return;

	VectorSubtract( vr_controllerOrientation, vr_lastControllerOrientation, angle );
}

void VR_GetOrientation( vec3_t angle )
{
	if( !hmd ) return;

	if( stale && VR_GetSensorOrientation() ) stale = 0;

	VectorCopy( vr_orientation, angle );
}

void VR_GetOrientationDelta( vec3_t angle )
{
	if( !hmd ) return;

	if( stale && VR_GetSensorOrientation() ) stale = 0;

	VectorSubtract( vr_orientation, vr_lastOrientation, angle );
}

// unused as of 9/30/2013 - consider removing later
void VR_GetOrientationEMA( vec3_t angle )
{
	vec4_t quat;
	if( !hmd ) return;

	VR_GetOrientationEMAQuat( quat );
	QuatToEuler( quat, angle );
}

// returns head offset using X forward, Y left, Z up
int32_t VR_GetHeadOffset( vec3_t offset )
{
	if( !hmd || !vr_positiontracking->value ) return false;

	if( stale && VR_GetSensorOrientation() ) stale = 0;

	return ( hmd->getHeadOffset && hmd->getHeadOffset( offset ) );
}

int32_t VR_GetHMDPos( int32_t * xpos, int32_t * ypos )
{
	if( hmd && hmd->getHMDPos )
	{
		hmd->getHMDPos( xpos, ypos );
		return 1;
	}
	return 0;
}

int32_t VR_GetHMDResolution( int32_t * width, int32_t * height )
{
	if( hmd && hmd->getHMDResolution )
	{
		hmd->getHMDResolution( width, height );
		return 1;
	}
	return 0;
}

void VR_GetOrientationEMAQuat( vec4_t quat )
{
	float  bounce = ( ( vr_hud_bounce->value == VR_HUD_BOUNCE_NONE ) ? 0 : 1 ) / vr_hud_bounce_falloff->value;
	vec4_t t1, t2, zero = {1, 0, 0, 0};

	if( !hmd ) return;

	if( stale && VR_GetSensorOrientation() ) stale = 0;

	EulerToQuat( vr_orientation, t1 );
	QuatCopy( vr_smoothOrientation, t2 );
	// fast conjugate
	t2[0] = -t2[0];
	QuatMultiply( t2, t1, t1 );
	SlerpQuat( zero, t1, bounce, quat );
}

void VR_FrameStart()
{
	if( !vr_enabled->value || !hmd ) return;

	// check and clamp cvars
	if( vr_aimmode->modified )
	{
		if( vr_aimmode->value < 0.0 )
			Cvar_SetInteger( "vr_aimmode", 0 );
		else if( vr_aimmode->value >= NUM_VR_AIMMODE )
			Cvar_SetInteger( "vr_aimmode", NUM_VR_AIMMODE - 1 );
		else
			Cvar_SetInteger( "vr_aimmode", (int32_t)vr_aimmode->value );
		VectorAdd( cl.bodyangles, cl.viewangles, cl.bodyangles );
		VectorSet( cl.aimdelta, 0, 0, 0 );
		VectorSet( cl.viewangles, 0, 0, 0 );
		vr_aimmode->modified = false;
	}

	if( vr_aimmode_deadzone_pitch->modified )
	{
		if( vr_aimmode_deadzone_pitch->value < 0.0 )
			Cvar_SetInteger( "vr_aimmode_deadzone_pitch", 0 );
		else if( vr_aimmode_deadzone_pitch->value > 360 )
			Cvar_SetInteger( "vr_aimmode_deadzone_pitch", 360 );
		else
			Cvar_SetInteger( "vr_aimmode_deadzone_pitch", (int32_t)vr_aimmode_deadzone_pitch->value );
		vr_aimmode_deadzone_pitch->modified = false;
	}

	if( vr_aimmode_deadzone_yaw->modified )
	{
		if( vr_aimmode_deadzone_yaw->value < 0.0 )
			Cvar_SetInteger( "vr_aimmode_deadzone_yaw", 0 );
		else if( vr_aimmode_deadzone_yaw->value > 360 )
			Cvar_SetInteger( "vr_aimmode_deadzone_yaw", 360 );
		else
			Cvar_SetInteger( "vr_aimmode_deadzone_yaw", (int32_t)vr_aimmode_deadzone_yaw->value );
		vr_aimmode_deadzone_yaw->modified = false;
	}

	if( vr_hud_deadzone_yaw->modified )
	{
		if( vr_hud_deadzone_yaw->value < 0.0 )
			Cvar_SetInteger( "vr_hud_deadzone_yaw", 0 );
		else if( vr_hud_deadzone_yaw->value > 360 )
			Cvar_SetInteger( "vr_hud_deadzone_yaw", 360 );
		else
			Cvar_SetInteger( "vr_hud_deadzone_yaw", (int32_t)vr_hud_deadzone_yaw->value );
		vr_hud_deadzone_yaw->modified = false;
	}

	if( vr_controllermode->modified )
	{
		if( vr_controllermode->value < 0.0 )
			Cvar_SetInteger( "vr_controllermode", 0 );
		else if( vr_controllermode->value >= NUM_VR_CONTROLLERMODE )
			Cvar_SetInteger( "vr_controllermode", NUM_VR_CONTROLLERMODE - 1 );
		else
			Cvar_SetInteger( "vr_controllermode", (int32_t)vr_controllermode->value );
		vr_controllermode->modified = false;
	}

	if( hand->value != 2 && (int32_t)vr_controllermode->value != 0 && hand->value != (int32_t)vr_controllermode->value - 1 )
	{
		hand->value = (int32_t)vr_controllermode->value - 1;
		Cvar_SetInteger( "hand", hand->value );
	}

	if( vr_aimlaser->modified )
	{
		Cvar_SetInteger( "vr_aimlaser", !!vr_aimlaser->value );
		vr_aimlaser->modified = false;
	}

	if( vr_hud_bounce->modified )
	{
		if( vr_hud_bounce->value < 0.0 )
			Cvar_SetInteger( "vr_hud_bounce", 0 );
		else if( vr_hud_bounce->value > NUM_HUD_BOUNCE_MODES )
			Cvar_SetInteger( "vr_hud_bounce", NUM_HUD_BOUNCE_MODES - 1 );
		vr_hud_bounce->modified = false;
	}

	if( vr_hud_bounce_falloff->modified )
	{
		if( vr_hud_bounce_falloff->value < 1.0 )
			Cvar_SetInteger( "vr_hud_bounce_falloff", 1 );
		else if( vr_hud_bounce_falloff->value > 60 )
			Cvar_SetInteger( "vr_hud_bounce_falloff", 60 );
		vr_hud_bounce_falloff->modified = false;
	}

	if( vr_autofov_scale->modified )
	{
		if( vr_autofov_scale->value > 2.0 ) Cvar_Set( "vr_autofov_scale", "2.0" );
		if( vr_autofov_scale->value < 0.5 ) Cvar_Set( "vr_autofov_scale", "0.5" );
		vr_autofov_scale->modified = false;
	}

	if( vr_prediction->modified )
	{
		if( vr_prediction->value < 0.0 )
			Cvar_Set( "vr_prediction", "0.0" );
		else if( vr_prediction->value > 75.0f )
			Cvar_Set( "vr_prediction", "75.0" );
		vr_prediction->modified = false;
		if( hmd->setPrediction ) hmd->setPrediction( vr_prediction->value );
	}

	if( vr_walkspeed->modified )
	{
		if( vr_walkspeed->value < 0.25 )
			Cvar_Set( "vr_walkspeed", "0.25" );
		else if( vr_walkspeed->value > 1.5 )
			Cvar_Set( "vr_walkspeed", "1.5" );

		vr_walkspeed->modified = false;
	}

	hmd->frameStart();

	// 	only flag as stale after a frame is rendered
	stale           = 1;
	controllerstale = 1;
}

void VR_FrameEnd()
{
	if( hmd && hmd->frameEnd ) hmd->frameEnd();
}

void VR_ResetOrientation( void )
{
	int    result;
	vec3_t angle;
	if( hmd && hmd->resetOrientation )
	{
		hmd->resetOrientation();
		VR_GetSensorOrientation();
		if( vr_aimmode->value == VR_AIMMODE_DECOUPLED )
		{
			vr_orientation_deltayaw += vr_lastOrientation[YAW] - vr_orientation[YAW];
			VectorSet( cl.aimdelta, 0, 0, 0 );
			VectorSet( cl.bodyangles, 0, 0, 0 );
			VectorSet( cl.viewangles, 0, 0, 0 );
		}
		else
		{
			vr_orientation_deltayaw = 0;
			VectorAdd( cl.bodyangles, cl.viewangles, cl.bodyangles );
			if( vr_comfortturn->value == 0 ) { VectorSet( cl.aimdelta, 0, 0, 0 ); }
			VectorSet( cl.viewangles, 0, 0, 0 );
		}
		VectorCopy( vr_orientation, vr_lastOrientation );

		result = VR_GetControllerOrientation( angle );
		if( result ) { VectorCopy( vr_controllerOrientation, vr_lastControllerOrientation ); }
		else
		{
			VectorSet( vr_controllerOrientation, 0, 0, 0 );
			VectorSet( vr_lastControllerOrientation, 0, 0, 0 );
		}
		QuatSet( vr_smoothSeries, 1, 0, 0, 0 );
		QuatSet( vr_doubleSmoothSeries, 1, 0, 0, 0 );
	}
	else
	{
		QuatSet( vr_smoothSeries, 1, 0, 0, 0 );
		QuatSet( vr_doubleSmoothSeries, 1, 0, 0, 0 );
		VectorSet( vr_orientation, 0, 0, 0 );
		VectorSet( vr_lastOrientation, 0, 0, 0 );
		VectorSet( vr_controllerOrientation, 0, 0, 0 );
		VectorSet( vr_lastControllerOrientation, 0, 0, 0 );
	}

	IN_CenterView();
}

void VR_ComfortTurn_Left( void )
{
	if( vr_comfortturn->value ) { vr_comfortturn_delta = vr_comfortturn->value; }
}

void VR_ComfortTurn_Right( void )
{
	if( vr_comfortturn->value ) { vr_comfortturn_delta = -vr_comfortturn->value; }
}

void VR_InputCommands()
{
	if( hmd && hmd->getInputCommands ) { hmd->getInputCommands(); }
}

void VR_IN_Move( usercmd_t * cmd )
{
	if( hmd && hmd->getControllerMoveInput ) { hmd->getControllerMoveInput( cmd ); }
}

// this is a little overkill on the abstraction, but the point of it was to try to make it
// possible to support other HMD's in the future without having to rewrite other portions of the engine.
// -dghost 08/17/13

int32_t VR_Enable()
{
	char    hmd_type[3];
	int32_t i;

	hmd = NULL;
	for( i = 1; i < NUM_HMD_TYPES; i++ )
	{
		Com_Printf( "VR: Checking for %s HMD...\n", hmd_names[i] );
		if( available_hmds[i].enable && available_hmds[i].enable() )
		{
			hmd = &available_hmds[i];
			break;
		}
	}

	if( !hmd )
	{
		Com_Printf( "VR: Error - no HMD attached.\n" );
		return 0;
	}

	Com_Printf( "VR: using %s HMD.\n", hmd_names[i] );

	if( hmd->setPrediction ) hmd->setPrediction( vr_prediction->value );
	vr_prediction->modified = false;

	VR_ResetOrientation();

	strncpy( hmd_type, va( "%i", hmd->type ), sizeof( hmd_type ) );
	Cvar_ForceSet( "vr_enabled", hmd_type );

	stale           = 1;
	controllerstale = 1;
	return 1;
}

void VR_Disable()
{
	hmd->disable();
	hmd = NULL;
	Cvar_ForceSet( "vr_enabled", "0" );
	Cvar_ForceSet( "vr_hmdstring", "VR Disabled" );
}

void VR_Teardown()
{
	int32_t i = 0;
	if( vr_enabled->value ) VR_Disable();

	for( i = 0; i < NUM_HMD_TYPES; i++ )
	{
		if( available_hmds[i].shutdown ) available_hmds[i].shutdown();
	}
}

// console command to enable Rift support
void VR_Enable_f( void )
{
	if( vr_enabled->value ) return;

	if( VR_Enable() ) { Cmd_ExecuteString( "vid_restart" ); }
}

// console command to disable Rift support
void VR_Disable_f( void )
{
	if( !vr_enabled->value ) return;

	VR_Disable();
	Cmd_ExecuteString( "vid_restart" );
}

// launch-time initialization for Rift support
void VR_Startup( void )
{
	int32_t i = 0;
	Com_Printf( "\n--------- VR Initialization ---------\n" );

	available_hmds[HMD_NONE] = hmd_none;
	available_hmds[HMD_RIFT] = hmd_rift;

	for( i = 0; i < NUM_HMD_TYPES; i++ )
	{
		if( available_hmds[i].init ) available_hmds[i].init();
	}

	vr_walkspeed              = Cvar_Get( "vr_walkspeed", "1.0", CVAR_CLIENT );
	vr_viewmove               = Cvar_Get( "vr_viewmove", "0", CVAR_CLIENT );
	vr_positiontracking       = Cvar_Get( "vr_positiontracking", "1", CVAR_CLIENT );
	vr_prediction             = Cvar_Get( "vr_prediction", "0", CVAR_CLIENT );
	vr_nosleep                = Cvar_Get( "vr_nosleep", "1", CVAR_CLIENT );
	vr_neckmodel_up           = Cvar_Get( "vr_neckmodel_up", "0.232", CVAR_CLIENT );
	vr_neckmodel_forward      = Cvar_Get( "vr_neckmodel_forward", "0.09", CVAR_CLIENT );
	vr_neckmodel              = Cvar_Get( "vr_neckmodel", "1", CVAR_CLIENT );
	vr_heightoffset           = Cvar_Get( "vr_heightoffset", "0.19", CVAR_CLIENT );
	vr_ipd                    = Cvar_Get( "vr_ipd", "64", CVAR_CLIENT );
	vr_hud_width              = Cvar_Get( "vr_hud_width", "800", CVAR_CLIENT );
	vr_hud_transparency       = Cvar_Get( "vr_hud_transparency", "1", CVAR_CLIENT );
	vr_hud_segments           = Cvar_Get( "vr_hud_segments", "15", CVAR_CLIENT );
	vr_hud_height             = Cvar_Get( "vr_hud_height", "600", CVAR_CLIENT );
	vr_hud_fov                = Cvar_Get( "vr_hud_fov", "58", CVAR_CLIENT );
	vr_hud_depth              = Cvar_Get( "vr_hud_depth", "1.0", CVAR_CLIENT );
	vr_hud_bounce_falloff     = Cvar_Get( "vr_hud_bounce_falloff", "15", CVAR_CLIENT );
	vr_hud_deadzone_yaw       = Cvar_Get( "vr_hud_deadzone_yaw", "20", CVAR_CLIENT );
	vr_hud_fixed              = Cvar_Get( "vr_hud_fixed", "1", CVAR_CLIENT );
	vr_hud_bounce             = Cvar_Get( "vr_hud_bounce", "2", CVAR_CLIENT );
	vr_hmdstring              = Cvar_Get( "vr_hmdstring", "VR Disabled", CVAR_NOSET );
	vr_force_resolution       = Cvar_Get( "vr_force_resolution", "0", CVAR_CLIENT );
	vr_force_fullscreen       = Cvar_Get( "vr_force_fullscreen", "1", CVAR_CLIENT );
	vr_supersampling          = Cvar_Get( "vr_supersampling", "1.0", CVAR_CLIENT );
	vr_enabled                = Cvar_Get( "vr_enabled", "0", CVAR_NOSET | CVAR_USERINFO );
	vr_aimlaser               = Cvar_Get( "vr_aimlaser", "0", CVAR_CLIENT );
	vr_chromatic              = Cvar_Get( "vr_chromatic", "1", CVAR_CLIENT );
	vr_autoipd                = Cvar_Get( "vr_autoipd", "1", CVAR_CLIENT );
	vr_autofov_scale          = Cvar_Get( "vr_autofov_scale", "1.0", CVAR_CLIENT );
	vr_autofov                = Cvar_Get( "vr_autofov", "1", CVAR_CLIENT );
	vr_autoenable             = Cvar_Get( "vr_autoenable", "1", CVAR_CLIENT );
	vr_aimmode_deadzone_yaw   = Cvar_Get( "vr_aimmode_deadzone_yaw", "40", CVAR_CLIENT );
	vr_aimmode_deadzone_pitch = Cvar_Get( "vr_aimmode_deadzone_pitch", "60", CVAR_CLIENT );
	vr_aimmode                = Cvar_Get( "vr_aimmode", "8", CVAR_CLIENT );
	vr_controllermode         = Cvar_Get( "vr_controllermode", "1", CVAR_CLIENT );
	vr_comfortturn            = Cvar_Get( "vr_comfortturn", "45", CVAR_CLIENT );
	vr_autocrouch             = Cvar_Get( "vr_autocrouch", "0", CVAR_CLIENT );
	Cmd_AddCommand( "vr_reset_home", VR_ResetOrientation );
	Cmd_AddCommand( "comfortturn_left", VR_ComfortTurn_Left );
	Cmd_AddCommand( "comfortturn_right", VR_ComfortTurn_Right );
	Cmd_AddCommand( "vr_disable", VR_Disable_f );
	Cmd_AddCommand( "vr_enable", VR_Enable_f );

	vr_orientation_deltayaw = 0;
	vr_comfortturn_delta    = 0;

	if( vr_autoenable->value ) VR_Enable();
	Com_Printf( "-------------------------------------\n" );
}
