#include "header/vr_ovr.h"
#include <math.h>

float ovrRadToDegree( float rad ) { return rad * MATH_FLOAT_RADTODEGREEFACTOR; }
float ovrDegreeToRad( float deg ) { return deg * MATH_FLOAT_DEGREETORADFACTOR; }

float ovrFovPort_GetHorizontalFovRadians( ovrFovPort fov ) { return atanf( fov.LeftTan ) + atanf( fov.RightTan ); }
float ovrFovPort_GetVerticalFovRadians( ovrFovPort fov ) { return atanf( fov.UpTan ) + atanf( fov.DownTan ); }

float ovrMatrix4f_Distance( ovrMatrix4f m1, ovrMatrix4f m2 )
{
	float result = 0;
	int   i, j;
	for( i = 0; i < 4; ++i )
		for( j = 0; j < 4; ++j ) result += fabs( m1.M[i][j] - m2.M[i][j] );
	return result;
}

ovrMatrix4f ovrMatrix4f_Identity()
{
	ovrMatrix4f result = {};
	result.M[0][0]     = 1.0f;
	result.M[1][1]     = 1.0f;
	result.M[2][2]     = 1.0f;
	result.M[3][3]     = 1.0f;
	return result;
}

ovrQuatf ovrQuatf_fromMatrix4f( ovrMatrix4f m )
{
	ovrQuatf result = {};
	float    trace  = m.M[0][0] + m.M[1][1] + m.M[2][2];

	if( trace > 0.0f )
	{
		trace    = sqrt( trace + 1.0f ) * 2.0f;
		result.w = 0.25f * trace;
		result.x = ( m.M[2][1] - m.M[1][2] ) / trace;
		result.y = ( m.M[0][2] - m.M[2][0] ) / trace;
		result.z = ( m.M[1][0] - m.M[0][1] ) / trace;
	}
	else if( ( m.M[0][0] > m.M[1][1] ) && ( m.M[0][0] > m.M[2][2] ) )
	{
		trace    = sqrt( 1.0f + m.M[0][0] - m.M[1][1] - m.M[2][2] ) * 2.0;
		result.w = ( m.M[2][1] - m.M[1][2] ) / trace;
		result.x = 0.25f * trace;
		result.y = ( m.M[0][1] - m.M[1][0] ) / trace;
		result.z = ( m.M[2][0] - m.M[0][2] ) / trace;
	}
	else if( m.M[1][1] > m.M[2][2] )
	{
		trace    = sqrt( 1.0f + m.M[1][1] - m.M[0][0] - m.M[2][2] ) * 2.0;
		result.w = ( m.M[0][2] - m.M[2][0] ) / trace;
		result.x = ( m.M[0][1] - m.M[1][0] ) / trace;
		result.y = 0.25f * trace;
		result.z = ( m.M[1][2] - m.M[2][1] ) / trace;
	}
	else
	{
		trace    = sqrt( 1.0f + m.M[2][2] - m.M[0][0] - m.M[1][1] ) * 2.0;
		result.w = ( m.M[1][0] - m.M[0][1] ) / trace;
		result.x = ( m.M[0][2] - m.M[2][0] ) / trace;
		result.y = ( m.M[1][2] - m.M[2][1] ) / trace;
		result.z = 0.25f * trace;
	}
	return result;
}

ovrVector3f ovrMatrix4f_GetTranslation( ovrMatrix4f m )
{
	ovrVector3f result = {};
	result.x           = m.M[0][3];
	result.y           = m.M[1][3];
	result.z           = m.M[2][3];
	return result;
}

#ifndef Q2VR_OCULUS
const char * ovr_GetVersionString() { return OVR_VERSION_STRING; }

#ifdef _WIN32
#include <windows.h>
double ovr_GetTimeInSeconds()
{
	static BOOL     hasHighRes = FALSE;
	static LONGLONG frequency  = 0;
	static LONGLONG startTime  = 0;
	LARGE_INTEGER   li;

	if( frequency == 0 )
	{
		hasHighRes = QueryPerformanceFrequency( &li );
		frequency = li.QuadPart;

		if( hasHighRes )
		{
			QueryPerformanceCounter( &li );
			startTime = li.QuadPart;
		}
		else
		{
			frequency = 1000;
			startTime = timeGetTime();
		}
	}

	if( hasHighRes )
		QueryPerformanceCounter( &li );
	else
		li.QuadPart = timeGetTime();

	return ( li.QuadPart - startTime ) / (double)( frequency );
}
#else
#include <sys/time.h>
#include <unistd.h>
#if defined( _POSIX_VERSION ) && ( _POSIX_VERSION >= 199309L )
#include <time.h>
#if defined( CLOCK_MONOTONIC_RAW )
#define Q2VR_CLOCK CLOCK_MONOTONIC_RAW
#elif defined( CLOCK_MONOTONIC )
#define Q2VR_CLOCK CLOCK_MONOTONIC
#else
#define Q2VR_CLOCK CLOCK_REALTIME
#endif
#elif defined( __STDC_VERSION__ ) && ( __STDC_VERSION__ >= 201112L )
#include <time.h>
#define Q2VR_CLOCK TIME_UTC
#define clock_gettime( c, t ) ( timespec_get( t, c ) ? 0 : 1 )
#else
#define Q2VR_CLOCK
#define clock_gettime( c, t ) 1
#define timespec timeval
#endif
double ovr_GetTimeInSeconds()
{
	static ovrBool   started    = ovrFalse;
	static ovrBool   hasHighRes = ovrFalse;
	static long long startTime  = 0;
	static long long frequency  = 0;
	struct timeval   tv;
	struct timespec  ts;

	if( started == ovrFalse )
	{
		if( clock_gettime( Q2VR_CLOCK, &ts ) == 0 )
		{
			hasHighRes = ovrTrue;
			frequency  = 1000000000;
			startTime  = ts.tv_nsec + ts.tv_sec * frequency;
		}
		else
		{
			gettimeofday( &tv, NULL );
			frequency = 1000000;
			startTime = tv.tv_usec + tv.tv_sec * frequency;
		}
		started = ovrTrue;
	}

	if( hasHighRes == ovrTrue )
	{
		clock_gettime( Q2VR_CLOCK, &ts );
		return ( ts.tv_nsec + ts.tv_sec * frequency - startTime ) / (double)( frequency );
	}
	else
	{
		gettimeofday( &tv, NULL );
		return ( tv.tv_usec + tv.tv_sec * frequency - startTime ) / (double)( frequency );
	}
}
#endif
#endif
