#ifndef __VR_H
#define __HR_H

#include "../../common/header/common.h"
#include <stdint.h>

#define PLAYER_HEIGHT_UNITS 56.0f
#define PLAYER_HEIGHT_M 1.75f

typedef enum vy_eye_t_
{
	EYE_LEFT  = 0,
	EYE_RIGHT = 1,
	NUM_EYES  = 2,
} vr_eye_t;

typedef struct eye_order_t_
{
	vr_eye_t eyes[NUM_EYES];
} eye_order_t;

typedef enum hmd_t_
{
	HMD_NONE,
	HMD_RIFT,
	NUM_HMD_TYPES,
} hmd_t;

extern const char * hmd_names[];

typedef struct hmd_interface_t_
{
	hmd_t type;
	int32_t ( *init )();
	void ( *shutdown )();
	int32_t ( *enable )();
	void ( *disable )();
	void ( *frameStart )();
	void ( *frameEnd )();
	void ( *resetOrientation )();
	int32_t ( *getOrientation )( float euler[3] );
	int32_t ( *getHeadOffset )( float offset[3] );
	int32_t ( *setPrediction )( float timeInMs );
	void ( *getHMDPos )( int32_t * xpos, int32_t * ypos );
	void ( *getHMDResolution )( int32_t * width, int32_t * height );
	int32_t ( *getControllerOrientation )( float euler[3] );
	void ( *getInputCommands )();
	void ( *getControllerMoveInput )();
} hmd_interface_t;

extern cvar_t * vr_enabled;
extern cvar_t * vr_autoenable;
extern cvar_t * vr_autofov;
extern cvar_t * vr_autofov_scale;
extern cvar_t * vr_autoipd;
extern cvar_t * vr_autocrouch;
extern cvar_t * vr_ipd;
extern cvar_t * vr_hud_fov;
extern cvar_t * vr_hud_depth;
extern cvar_t * vr_hud_segments;
extern cvar_t * vr_hud_transparency;
extern cvar_t * vr_hud_width;
extern cvar_t * vr_hud_height;
extern cvar_t * vr_chromatic;
extern cvar_t * vr_aimlaser;
extern cvar_t * vr_aimmode;
extern cvar_t * vr_aimmode_deadzone_yaw;
extern cvar_t * vr_aimmode_deadzone_pitch;
extern cvar_t * vr_controllermode;
extern cvar_t * vr_comfortturn;
extern cvar_t * vr_viewmove;
extern cvar_t * vr_hud_bounce;
extern cvar_t * vr_hud_bounce_falloff;
extern cvar_t * vr_hud_deadzone_yaw;
extern cvar_t * vr_hud_fixed;
extern cvar_t * vr_nosleep;
extern cvar_t * vr_neckmodel;
extern cvar_t * vr_neckmodel_up;
extern cvar_t * vr_neckmodel_forward;
extern cvar_t * vr_heightoffset;
extern cvar_t * vr_walkspeed;
extern cvar_t * vr_force_fullscreen;
extern cvar_t * vr_force_resolution;
extern cvar_t * vr_supersampling;

typedef enum vr_aimmode_t_
{
	VR_AIMMODE_DISABLE,
	VR_AIMMODE_HEAD_MYAW,
	VR_AIMMODE_HEAD_MYAW_MPITCH,
	VR_AIMMODE_MOUSE_MYAW,
	VR_AIMMODE_MOUSE_MYAW_MPITCH,
	VR_AIMMODE_TF2_MODE2,
	VR_AIMMODE_TF2_MODE3,
	VR_AIMMODE_TF2_MODE4,
	VR_AIMMODE_DECOUPLED,
	NUM_VR_AIMMODE,
} vr_aimmode_t;

typedef enum vr_controllermode_t_
{
	VR_CONTROLLERMODE_DISABLE,
	VR_CONTROLLERMODE_RIGHTHAND,
	VR_CONTROLLERMODE_LEFTHAND,
	NUM_VR_CONTROLLERMODE,
} vr_controllermode_t;

typedef enum vr_bounce_mode_t_
{
	VR_HUD_BOUNCE_NONE,
	VR_HUD_BOUNCE_SES,
	VR_HUD_BOUNCE_LES,
	NUM_HUD_BOUNCE_MODES,
} vr_bounce_mode_t;

extern float vr_orientation_deltayaw;
extern float vr_comfortturn_delta;

void    VR_Startup();
void    VR_Teardown();
int32_t VR_Enable();
void    VR_Disable();
void    VR_FrameStart();
void    VR_FrameEnd();
int32_t VR_GetControllerOrientation( vec3_t angle );
void    VR_GetControllerOrientationDelta( vec3_t angle );
void    VR_GetOrientation( vec3_t angle );
void    VR_GetOrientationDelta( vec3_t angle );
void    VR_GetOrientationEMA( vec3_t angle );
void    VR_GetOrientationEMAQuat( vec4_t quat );
int32_t VR_GetHeadOffset( vec3_t offset );
void    VR_ResetOrientation();
void    VR_ComfortTurn_Left();
void    VR_ComfortTurn_Right();
void    VR_InputCommands();
void    VR_IN_Move( usercmd_t * cmd );
int32_t VR_GetHMDPos( int32_t * xpos, int32_t * ypos );
int32_t VR_GetHMDResolution( int32_t * width, int32_t * height );

#endif

#ifndef __VR_RIFT_H
#define __VR_RIFT_H

typedef struct
{
	char label[64];
	char serialnumber[24];
} riftname_t;

typedef struct
{
	qboolean withinFrame;
	float    cameraYaw;
	qboolean positionTracked;
	qboolean hasPositionLock;
} rift_render_export_t;

extern cvar_t *        vr_rift_debug;
extern cvar_t *        vr_rift_prediction;
extern cvar_t *        vr_rift_maxfov;
extern cvar_t *        vr_rift_autoprediction;
extern cvar_t *        vr_rift_trackingloss;
extern cvar_t *        vr_rift_dk2_color_hack;
extern hmd_interface_t hmd_rift;

#endif
