#ifndef CL_VR_OVR_H
#define CL_VR_OVR_H

#ifdef Q2VR_OCULUS
#include <OVR_CAPI.h>
#else
#include <stdint.h>

#if defined( __GNUC__ ) || defined( __clang__ )
#define OVR_ALIGNAS( n ) __attribute__( ( aligned( n ) ) )
#elif defined( _MSC_VER ) || defined( __INTEL_COMPILER )
#define OVR_ALIGNAS( n ) __declspec( align( n ) )
#elif defined( __CC_ARM )
#define OVR_ALIGNAS( n ) __align( n )
#else
#define OVR_ALIGNAS( n )
#endif

#ifdef _WIN32
#define OVR_CDECL __cdecl
#else
#define OVR_CDECL
#endif

#if defined( __WORDSIZE )
#define OVR_PTR_SIZE ( ( __WORDSIZE ) / 8 )
#elif defined( _WIN64 ) || defined( __LP64__ ) || defined( _LP64 ) || defined( _M_IA64 ) || defined( __ia64__ ) || defined( __arch64__ ) || defined( __64BIT__ ) || defined( __Ptr_Is_64 )
#define OVR_PTR_SIZE 8
#elif defined( __CC_ARM ) && ( __sizeof_ptr == 8 )
#define OVR_PTR_SIZE 8
#else
#define OVR_PTR_SIZE 4
#endif

#define OVRSTRINGIZEIMPL( x ) #x
#define OVR_STRINGIZE( x ) OVRSTRINGIZEIMPL( x )

#define OVR_SUCCESS( result ) ( result >= 0 )
#define OVR_UNQUALIFIED_SUCCESS( result ) ( result == ovrSuccess )
#define OVR_FAILURE( result ) ( !OVR_SUCCESS( result ) )

#define OVR_UNUSED_STRUCT_PAD( padName, size ) char padName[size]

#if OVR_PTR_SIZE == 8
#define OVR_ON32( x )
#define OVR_ON64( x ) x
#elif OVR_PTR_SIZE == 4
#define OVR_ON32( x ) x
#define OVR_ON64( x )
#else
#define OVR_ON32( x )
#define OVR_ON64( x )
#endif

#define OVR_MAJOR_VERSION 1
#define OVR_MINOR_VERSION 4
#define OVR_PATCH_VERSION 0
#define OVR_BUILD_NUMBER 0

#define OVR_VERSION_STRING OVR_STRINGIZE( OVR_MAJOR_VERSION.OVR_MINOR_VERSION.OVR_PATCH_VERSION )
#define OVR_DETAILED_VERSION_STRING OVR_STRINGIZE( OVR_MAJOR_VERSION.OVR_MINOR_VERSION.OVR_PATCH_VERSION.OVR_BUILD_NUMBER )

typedef enum ovrButton_
{
	ovrButton_A         = 0x00000001,
	ovrButton_B         = 0x00000002,
	ovrButton_RThumb    = 0x00000004,
	ovrButton_RShoulder = 0x00000008,
	ovrButton_RMask     = ovrButton_A | ovrButton_B | ovrButton_RThumb | ovrButton_RShoulder,

	ovrButton_X         = 0x00000100,
	ovrButton_Y         = 0x00000200,
	ovrButton_LThumb    = 0x00000400,
	ovrButton_LShoulder = 0x00000800,
	ovrButton_LMask     = ovrButton_X | ovrButton_Y | ovrButton_LThumb | ovrButton_LShoulder,

	ovrButton_Up    = 0x00010000,
	ovrButton_Down  = 0x00020000,
	ovrButton_Left  = 0x00040000,
	ovrButton_Right = 0x00080000,

	ovrButton_Enter = 0x00100000,
	ovrButton_Back  = 0x00200000,

	ovrButton_VolUp   = 0x00400000,
	ovrButton_VolDown = 0x00800000,
	ovrButton_Home    = 0x01000000,
	ovrButton_Private = ovrButton_VolUp | ovrButton_VolDown | ovrButton_Home,

	ovrButton_EnumSize = 0x7fffffff,
} ovrButton;

typedef enum ovrControllerType_
{
	ovrControllerType_None     = 0x00,
	ovrControllerType_LTouch   = 0x01,
	ovrControllerType_RTouch   = 0x02,
	ovrControllerType_Touch    = 0x03,
	ovrControllerType_Remote   = 0x04,
	ovrControllerType_XBox     = 0x10,
	ovrControllerType_Active   = 0xff,
	ovrControllerType_EnumSize = 0x7fffffff,
} ovrControllerType;

typedef enum ovrErrorType_
{
	ovrError_MemoryAllocationFailure   = -1000,
	ovrError_SocketCreationFailure     = -1001,
	ovrError_InvalidSession            = -1002,
	ovrError_timeout                   = -1003,
	ovrError_NotInitialized            = -1004,
	ovrError_InvalidParameter          = -1005,
	ovrError_ServiceError              = -1006,
	ovrError_NoHmd                     = -1007,
	ovrError_Unsupported               = -1009,
	ovrError_DeviceUnavailable         = -1010,
	ovrError_InvalidHeadsetOrientation = -1011,
	ovrError_ClientSkippedDestroy      = -1012,
	ovrError_ClientSkippedShutdown     = -1013,
	ovrError_ServiceDeadlockDetected   = -1014,

	ovrError_AudioReservedBegin  = -2000,
	ovrError_AudioDeviceNotFound = -2001,
	ovrError_AudioComError       = -2002,
	ovrError_AudioReservedEnd    = -2999,

	ovrError_Initialize                 = -3000,
	ovrError_LibLoad                    = -3001,
	ovrError_LibVersion                 = -3002,
	ovrError_ServiceConnection          = -3003,
	ovrError_ServiceVersion             = -3004,
	ovrError_IncompatibleOS             = -3005,
	ovrError_DisplayInit                = -3006,
	ovrError_ServerStart                = -3007,
	ovrError_Reinitialization           = -3008,
	ovrError_MismatchedAdapters         = -3009,
	ovrError_LeakingResources           = -3010,
	ovrError_ClientVersion              = -3011,
	ovrError_OutOfDateOS                = -3012,
	ovrError_OutOfDateGfxDriver         = -3013,
	ovrError_IncompatibleGPU            = -3014,
	ovrError_NoValidVRDisplaySystem     = -3015,
	ovrError_Obsolete                   = -3016,
	ovrError_DisabledOrDefaultAdapter   = -3017,
	ovrError_HybridGraphicsNotSupported = -3018,
	ovrError_DisplayManagerInit         = -3019,
	ovrError_TrackerDriverInit          = -3020,

	ovrError_RuntimeException = -7000,
} ovrErrorType;

typedef enum ovrEyeType_
{
	ovrEye_Left     = 0,
	ovrEye_Right    = 1,
	ovrEye_Count    = 2,
	ovreye_EnumSize = 0x7fffffff,
} ovrEyeType;

typedef enum ovrHandType_
{
	ovrHand_Left     = 0,
	ovrHand_Right    = 1,
	ovrHand_Count    = 2,
	ovrHand_EnumSize = 0x7fffffff,
} ovrHandType;

typedef enum ovrHmdType_
{
	ovrHmd_None     = 0,
	ovrHmd_DK1      = 3,
	ovrHmd_DKHD     = 4,
	ovrHmd_DK2      = 6,
	ovrHmd_CB       = 8,
	ovrHmd_Other    = 9,
	ovrHmd_E3_2015  = 10,
	ovrHmd_ES06     = 11,
	ovrHmd_ES09     = 12,
	ovrHmd_ES11     = 13,
	ovrHmd_CV1      = 14,
	ovrHmd_EnumSize = 0x7fffffff,
} ovrHmdType;

typedef enum ovrStatusBits_
{
	ovrStatus_OrientationTracked = 0x0001,
	ovrStatus_PositionTracked    = 0x0002,
	ovrStatus_EnumSize           = 0x7fffffff,
} ovrStatusBits;

typedef enum ovrSucessType_
{
	ovrSuccess = 0,

	ovrSuccess_Notvisible = 1000,

	ovrSuccess_HMDFirmwareMismatch        = 4100,
	ovrSuccess_TrackerFirmwareMismatch    = 4101,
	ovrSuccess_ControllerFirmwareMismatch = 4104,
	ovrSuccess_TrackerDriverNotFound      = 4105,
} ovrSuccessType;

typedef enum ovrTrackerFlags_
{
	ovrTracker_Connected   = 0x0020,
	ovrTracker_PoseTracked = 0x0004,
} ovrTrackerFlags;

typedef enum ovrTrackingCaps_
{
	ovrTrackingCap_Orientation      = 0x0010,
	ovrTrackingCap_MagYawCorrection = 0x0020,
	ovrTrackingCap_Position         = 0x0040,
	ovrTrackingCap_EnumSize         = 0x7fffffff,
} ovrTrackingCaps;

typedef enum ovrTrackingOrigin_
{
	ovrTrackingOrigin_EyeLevel   = 0,
	ovrTrackingOrigin_FloorLevel = 1,
	ovrTrackingOrigin_Count      = 2,
	ovrTrackingOrigin_EnumSize   = 0x7fffffff,
} ovrTrackingOrigin;

typedef char ovrBool;
#define ovrFalse 0
#define ovrTrue 1

typedef int32_t ovrResult;

typedef void( OVR_CDECL * ovrLogCallback )( uintptr_t userData, int level, const char * message );

typedef struct OVR_ALIGNAS( 4 ) ovrFovPort_
{
	float UpTan;
	float DownTan;
	float LeftTan;
	float RightTan;
} ovrFovPort;

typedef struct OVR_ALIGNAS( OVR_PTR_SIZE ) ovrGraphicsLuid_
{
	char Reserved[8];
} ovrGraphicsLuid;

typedef struct OVR_ALIGNAS( 4 ) ovrMatrix4f_
{
	float M[4][4];
} ovrMatrix4f;

typedef struct OVR_ALIGNAS( 4 ) ovrQuatf_
{
	float x, y, z, w;
} ovrQuatf;

typedef struct OVR_ALIGNAS( 4 ) ovrSizei_
{
	int w, h;
} ovrSizei;

typedef struct OVR_ALIGNAS( OVR_PTR_SIZE ) ovrTrackerDesc_
{
	float FrustumHFovInRadians;
	float FrustumVFovInRadians;
	float FrustumNearZInMeters;
	float FrustumFarZInMeters;
} ovrTrackerDesc;

typedef struct OVR_ALIGNAS( 4 ) ovrVector2f_
{
	float x, y;
} ovrVector2f;

typedef struct OVR_ALIGNAS( 4 ) ovrVector2i_
{
	int x, y;
} ovrVector2i;

typedef struct OVR_ALIGNAS( 4 ) ovrVector3f_
{
	float x, y, z;
} ovrVector3f;

// 1

typedef struct OVR_ALIGNAS( 8 ) ovrDetectResult_
{
	ovrBool IsOculusServiceRunning;
	ovrBool IsOculusHMDConnected;
	OVR_UNUSED_STRUCT_PAD( pad0, 6 );
} ovrDetectResult;

typedef struct ovrErrorInfo_
{
	ovrResult Result;
	char      ErrorString[512];
} ovrErrorInfo;

typedef struct OVR_ALIGNAS( 8 ) ovrInitParams_
{
	uint32_t       Flags;
	uint32_t       RequestedMinorVersion;
	ovrLogCallback LogCallback;
	uintptr_t      UserData;
	uint32_t       ConnectionTimeoutMS;
	OVR_ON64( OVR_UNUSED_STRUCT_PAD( pad0, 4 ) );
} ovrInitParams;

typedef struct ovrSessionStatus_
{
	ovrBool IsVisible;
	ovrBool HmdPresent;
	ovrBool HmdMounted;
	ovrBool DisplayLost;
	ovrBool ShouldQuit;
	ovrBool ShouldRecenter;
} ovrSessionStatus;

// 2

typedef struct ovrInputState_
{
	double            TimeInSeconds;
	unsigned int      Buttons;
	unsigned int      Touches;
	float             IndexTrigger[ovrHand_Count];
	float             HandTrigger[ovrHand_Count];
	ovrVector2f       Thumbstick[ovrHand_Count];
	ovrControllerType ControllerType;
} ovrInputState;

typedef struct OVR_ALIGNAS( 4 ) ovrPosef_
{
	ovrQuatf    Orientation;
	ovrVector3f Position;
} ovrPosef;

typedef struct OVR_ALIGNAS( 8 ) ovrPoseStatef_
{
	ovrPosef    ThePose;
	ovrVector3f AngularVelocity;
	ovrVector3f LinearVelocity;
	ovrVector3f AngularAcceleration;
	ovrVector3f LinearAcceleration;
	OVR_UNUSED_STRUCT_PAD( pad0, 4 );
	double TimeInSeconds;
} ovrPoseStatef;

typedef struct OVR_ALIGNAS( 4 ) ovrRecti_
{
	ovrVector2i Pos;
	ovrSizei    Size;
} ovrRecti;

typedef struct OVR_ALIGNAS( 8 ) ovrTrackingState_
{
	ovrPoseStatef HeadPose;
	unsigned int  StatusFlags;
	ovrPoseStatef HandPoses[2];
	unsigned int  HandStatusFlags[2];
	ovrPosef      CalibratedOrigin;
} ovrTrackingState;

// 3

typedef struct OVR_ALIGNAS( OVR_PTR_SIZE ) ovrHmdDesc_
{
	ovrHmdType Type;
	OVR_ON64( OVR_UNUSED_STRUCT_PAD( pad0, 4 ) );
	char         ProductName[64];
	char         Manufacturer[64];
	short        VendorId;
	short        ProductId;
	char         SerialNumber[24];
	short        FirmwareMajor;
	short        FirmwareMinor;
	unsigned int AvailableHmdCaps;
	unsigned int DefaultHmdCaps;
	unsigned int AvailableTrackingCaps;
	unsigned int DefaultTrackingCaps;
	ovrFovPort   DefaultEyeFov[ovrEye_Count];
	ovrFovPort   MaxEyeFov[ovrEye_Count];
	ovrSizei     Resolution;
	float        DisplayRefreshRate;
	OVR_ON64( OVR_UNUSED_STRUCT_PAD( pad1, 4 ) );
} ovrHmdDesc;

// 5

typedef struct OVR_ALIGNAS( 4 ) ovrEyeRenderDesc_
{
	ovrEyeType  Eye;
	ovrFovPort  Fov;
	ovrRecti    DistortedViewport;
	ovrVector2f PixelsPerTanAngleAtCenter;
	ovrPosef    HmdToEyePose;   // SDK 1.17+
	ovrVector3f HmdToEyeOffset; // SDK Legacy
} ovrEyeRenderDesc;

#ifndef HAS_OVR_HMD_STRUCT
typedef struct ovrHmdStruct_
{
} ovrHmdStruct;
#endif
typedef struct ovrHmdStruct_ * ovrSession;

extern ovrDetectResult  ovr_Detect( int timeout );
extern ovrEyeRenderDesc ovr_GetRenderDesc( ovrSession session, ovrEyeType eyeType, ovrFovPort fov );
extern ovrHmdDesc       ovr_GetHmdDesc( ovrSession session );
extern ovrResult        ovr_Create( ovrSession * pSession, ovrGraphicsLuid * pLuid );
extern ovrResult        ovr_GetInputState( ovrSession session, ovrControllerType controllerType, ovrInputState * inputState );
extern ovrResult        ovr_GetSessionStatus( ovrSession session, ovrSessionStatus * sessionStatus );
extern ovrResult        ovr_Initialize( const ovrInitParams * params );
extern ovrResult        ovr_RecenterTrackingOrigin( ovrSession session );
extern ovrResult        ovr_SetTrackingOriginType( ovrSession session, ovrTrackingOrigin origin );
extern ovrTrackerDesc   ovr_GetTrackerDesc( ovrSession session, unsigned int trackerDescIndex );
extern ovrTrackingState ovr_GetTrackingState( ovrSession session, double absTime, ovrBool latencyMarker );

extern const char * ovr_GetVersionString();
extern double       ovr_GetTimeInSeconds();
extern double       ovr_GetPredictedDisplayTime( ovrSession session, long long frameIndex );
extern unsigned int ovr_GetConnectedControllerTypes( ovrSession session );
extern void         ovr_Destroy( ovrSession session );
extern void         ovr_GetLastErrorInfo( ovrErrorInfo * errorInfo );
extern void         ovr_Shutdown();

#endif

#define MATH_DOUBLE_PI 3.14159265358979323846
#define MATH_DOUBLE_TWOPI ( 2 * MATH_DOUBLE_PI )
#define MATH_DOUBLE_PIOVER2 ( 0.5 * MATH_DOUBLE_PI )
#define MATH_DOUBLE_PIOVER4 ( 0.25 * MATH_DOUBLE_PI )

#define MATH_DOUBLE_RADTODEGREEFACTOR ( 360.0 / MATH_DOUBLE_TWOPI )
#define MATH_DOUBLE_DEGREETORADFACTOR ( MATH_DOUBLE_TWOPI / 360.0 )

#define MATH_FLOAT_PI (float)( MATH_DOUBLE_PI )
#define MATH_FLOAT_TWOPI (float)( MATH_DOUBLE_TWOPI )
#define MATH_FLOAT_PIOVER2 (float)( MATH_DOUBLE_PIOVER2 )
#define MATH_FLOAT_PIOVER4 (float)( MATH_DOUBLE_PIOVER4 )

#define MATH_FLOAT_RADTODEGREEFACTOR (float)( MATH_DOUBLE_RADTODEGREEFACTOR )
#define MATH_FLOAT_DEGREETORADFACTOR (float)( MATH_DOUBLE_DEGREETORADFACTOR )

extern ovrMatrix4f ovrMatrix4f_Identity();
extern ovrQuatf    ovrQuatf_fromMatrix4f( ovrMatrix4f m );
extern ovrVector3f ovrMatrix4f_GetTranslation( ovrMatrix4f m );

extern float ovrDegreeToRad( float deg );
extern float ovrFovPort_GetHorizontalFovRadians( ovrFovPort fov );
extern float ovrFovPort_GetVerticalFovRadians( ovrFovPort fov );
extern float ovrMatrix4f_Distance( ovrMatrix4f m1, ovrMatrix4f m2 );
extern float ovrRadToDegree( float rad );

#endif
